function mu = calc_bc_2dto3d(map2d, map3d, patch)
% Compute the point cloud Beltrami coefficients.
%
% If you use this code in your own work, please cite the following paper:
% [1] T. W. Meng, G. P.-T. Choi and L. M. Lui, 
%     "TEMPO: Feature-Endowed Teichmller Extremal Mappings of Point Clouds."
%     SIAM Journal on Imaging Sciences, 9(4), pp. 1922-1962, 2016.
%
% Copyright (c) 2015-2018, Gary Pui-Tung Choi
% https://scholar.harvard.edu/choi

[n,k] = size(patch.knnInd);

m2d_ = zeros(k, 2);
coeff_n = 6;
C = zeros(n, coeff_n*k);

for i=1:n
    m2d_(:,1) = map2d(patch.knnInd(i,:)',1) - map2d(i,1);
    m2d_(:,2) = map2d(patch.knnInd(i,:)',2) - map2d(i,2);
    
    norm = m2d_(:,1).^2+ m2d_(:,2).^2;
    maxnorm = max(norm);

    switch(patch.weight_id)
        case 1
            weight = exp(-sqrt(k)*norm/maxnorm); %gaussian
        case 2
            weight = exp(-sqrt(k)*norm/maxnorm)/k; %gaussian c1
        case 3
            weight = exp(-k*norm/maxnorm/2); %gaussian c2
    end
    weight(1) = 1;
    
    b = [ones(k,1), m2d_(:,1), m2d_(:,2), m2d_(:,1).^2, m2d_(:,1).*m2d_(:,2), m2d_(:,2).^2].';
    w = diag(weight);
    A = b * w * b.';
    B = b * w;
    
    c = A\B;
    C(i,:) = c(:);
    
end

knnInd = patch.knnInd;
ux = sum(C(:, 2:6:6*k-4) .* map3d(knnInd), 2);
vx = sum(C(:, 2:6:6*k-4) .* map3d(knnInd+n), 2);
wx = sum(C(:, 2:6:6*k-4) .* map3d(knnInd+2*n), 2);
uy = sum(C(:, 3:6:6*k-3) .* map3d(knnInd), 2);
vy = sum(C(:, 3:6:6*k-3) .* map3d(knnInd+n), 2);
wy = sum(C(:, 3:6:6*k-3) .* map3d(knnInd+2*n), 2);

E = ux.^2 + vx.^2 + wx.^2;
F = ux.*uy + vx.*vy + wx.*wy;
G = uy.^2 + vy.^2 + wy.^2;

D = E + G + 2*sqrt(E.*G - F.^2);
mu = ( E - G + 2i * F) ./ D;

mu = bdd_mu_localmesh(mu, map2d, map3d, patch);
end


function mu = bdd_mu_localmesh(mu, map2d, fz, patch)
if size(fz,2) < 3
    fz = [fz, zeros(size(fz, 1), 1)];
end

bdd = patch.bdd_index;
k = patch.k;
v_local = zeros(k, 2);

for i=1:size(bdd,1)
    index = bdd(i,1);
    knnInd_local = patch.knnInd(index,:)';
    
    v_local(:,1) = map2d(knnInd_local,1) - map2d(index,1);
    v_local(:,2) = map2d(knnInd_local,2) - map2d(index,2);
    
    fz_local = fz(knnInd_local, :);
    
    TR = delaunay(v_local(:,1), v_local(:,2));
    [row,~] = find(TR==1);
    oneringtri = TR(row,:);
    max_ind = max(oneringtri(:));
    mu_face = ones(size(oneringtri,1),1);
    mu_face(:,1) = bc_metric(v_local(1:max_ind, :)', oneringtri', fz_local(1:max_ind, :)', patch.dim);
    
    mu_local = 0;
    mu_local_count = 0;
    for j=1:size(oneringtri,1)
        if isnan(mu_face(j,1))
            continue;
        end
        mu_local = mu_local + mu_face(j,1);
        mu_local_count = mu_local_count + 1;
    end
    if mu_local_count==0
        if i>1
            last = bdd(i-1,1);
        else 
            last = bdd(size(bdd,1),1);
        end
        if i<size(bdd,1)
            next = bdd(i+1,1);
        else 
            next = bdd(1,1);
        end
        logical = ismember(knnInd_local(oneringtri), [last, bdd(i,1), next]);
        [accept_index, ~] = find(logical(:,1) & logical(:,2) & logical(:,3));
        if ~isempty(accept_index)
            mu(index,1) = mu_face(accept_index,1);
        else
            disp('bdd tri error!');
        end
    else
        mu(index,1) = mu_local ./ mu_local_count;
    end
end
end