function map = pc_disk_harmonic_map(v, bdy_index, bdd_pos, patch)
% Compute the point cloud harmonic map.
%
% If you use this code in your own work, please cite the following paper:
% [1] T. W. Meng, G. P.-T. Choi and L. M. Lui, 
%     "TEMPO: Feature-Endowed Teichmller Extremal Mappings of Point Clouds."
%     SIAM Journal on Imaging Sciences, 9(4), pp. 1922-1962, 2016.
%
% Copyright (c) 2015-2018, Gary Pui-Tung Choi
% https://scholar.harvard.edu/choi

n = size(v,1);

Ld = LdMatrix(v, bdy_index, patch.k, patch.weight_id, patch.order);

targetc = bdd_pos(:,1) + 1i*bdd_pos(:,2);
b = -Ld(:,bdy_index)*targetc;
b(bdy_index) = targetc;
Ld(bdy_index,:) = 0; Ld(:,bdy_index) = 0;
Ld(bdy_index,bdy_index) = diag(ones(length(bdy_index),1));

map = Ld\b;
map = [real(map) imag(map), zeros(n,1)];
end

function Ld = LdMatrix(v, bdd_i, k, weight_id, order)
n = size(v,1);
knnInd = knnsearch(v, v, 'K', k);

[coeff, powerc] = calc_coeff_power(order, 0,0);

v_local_1 = reshape(v(knnInd,1), [n,k]);
v_local_2 = reshape(v(knnInd,2), [n,k]);
v_local_3 = reshape(v(knnInd,3), [n,k]);

M = zeros(k,n);
N = zeros(k,n);
val = zeros(k,n);

[bxxcoeff, bxxpower] = calc_coeff_power(order, 2,0);
[byycoeff, byypower] = calc_coeff_power(order, 0,2);
bxx_ = bxxcoeff .* ( (0.^bxxpower(:,1)) .* (0.^bxxpower(:,2)));
byy_ = byycoeff .* ( (0.^byypower(:,1)) .* (0.^byypower(:,2)));

for j=1:n
    knn = zeros(k,3);
    % center to v(j,:)
    knn(:,1) = v_local_1(j,:)' - v(j,1);
    knn(:,2) = v_local_2(j,:)' - v(j,2);
    knn(:,3) = v_local_3(j,:)' - v(j,3);
    
    PCA = pca(knn,'Rows','complete');
    
    knn_local = knn * PCA;
    proj_k = knn - knn_local(:,3) * PCA(:,3)';
    knn_local(:,1:2) = proj_k * PCA(:,1:2);

    x_ = knn_local(:,1);
    y_ = knn_local(:,2);
    
    M(:,j) = j;
    N(:,j) = knnInd(j,:)';
    if isempty(find(bdd_i==j,1))
        norm = x_.^2+ y_.^2;
        maxnorm = max(norm);
        
        switch(weight_id)
            case 1
                weight = exp(-sqrt(k)*norm/maxnorm); %gaussian
            case 2
                weight = exp(-sqrt(k)*norm/maxnorm)/k; %gaussian c1
            case 3
                weight = exp(-k*norm/maxnorm/2); %gaussian c2
        end
        
        weight(1) = 1;
        
        A = zeros(length(coeff), length(coeff));
        B = zeros(length(coeff), k);
        for i=1:k
            b = coeff .* (x_(i).^powerc(:,1)) .* (y_(i).^powerc(:,2));
            B(:,i) = weight(i) * b;
            A = A + B(:,i) * b';
        end
        
        c = A\B;
        bx_ = [0;1;0;0;0;0];
    by_ = [0;0;1;0;0;0];
    bxx_ = [0;0;0;2;0;0];
    bxy_ = [0;0;0;0;1;0];
    byy_ = [0;0;0;0;0;2];
    D = [bx_, by_, bxx_, bxy_, byy_].' * c * knn_local(:,3);
    
    val(:,j) = c'*...
        (bx_*(2*D(1)^2*D(2)*D(4)-D(1)*D(3)-D(1)*D(2)^2*D(3)-D(1)*D(5)-D(1)^3*D(5))...
        +by_*(2*D(1)*D(2)^2*D(4)-D(2)*D(3)-D(2)^3*D(3)-D(2)*D(5)-D(1)^2*D(2)*D(5))...
        +bxx_*((1+D(1)^2+D(2)^2)*(1+D(2)^2))...
        +bxy_*((1+D(1)^2+D(2)^2)*(-2*D(1)*D(2)))...
        +byy_*((1+D(1)^2+D(2)^2)*(1+D(1)^2)));
    else
    end
end

Ld = sparse(M, N, val, n, n);
end

