function [map,map_mu] = pc_rectangular_conformal_map(v,bdy_index,corner)
% Compute the point cloud rectangular conformal mapping using the fast method in [1].
%
% Input:
% v: nv x 3 vertex coordinates of a disk-shaped point cloud
% bdy_index: nb x 1 vertex indices of the boundary, with anti-clockwise orientation
% (optional) corner: 4 x 1 vertex indices for the four corners of the rectangle, with anti-clockwise orientation
% 4 - 3
% |   |
% 1 - 2
%
% Output:
% map: nv x 2 vertex coordinates of the rectangular conformal parameterization
% map_mu: nv x 1 Beltrami coefficient 
%
% If you use this code in your own work, please cite the following paper:
% [1] T. W. Meng, G. P.-T. Choi and L. M. Lui, 
%     "TEMPO: Feature-Endowed Teichmller Extremal Mappings of Point Clouds."
%     SIAM Journal on Imaging Sciences, 9(4), pp. 1922-1962, 2016.
%
% Copyright (c) 2015-2018, Gary Pui-Tung Choi
% https://scholar.harvard.edu/choi

%% Initial setup

nv = length(v);
nb = length(bdy_index);

if nargin < 3
    % if the corner indices are not specified, pick 4 boundary points as corner
    corner = bdy_index([1, floor(nb/4), floor(nb/2), floor(3*nb/4)]);
end

if sum(ismember(corner,bdy_index)) ~= 4
    error('The input corner indices do not lie on the surface boundary!');
end

% rearrange the boundary indices
id = find(bdy_index == corner(1));
bdy_index = bdy_index([id:end,1:id-1]);

id1 = 1;
id2 = find(bdy_index == corner(2));
id3 = find(bdy_index == corner(3));
id4 = find(bdy_index == corner(4));

patch.k = 36;
patch.weight_id = 2;
patch.order = 2;
patch.dim = 3;
patch.bdd_index = bdy_index;

%% Step 1: Mapping the input mesh onto the unit disk

% disk bdy condition
diskbdy_x = zeros(nb,1);
diskbdy_y = zeros(nb,1);
bdd_len = 0;
for i = 1:nb-1
    bdd_len = bdd_len + norm(v(bdy_index(i+1),:)-v(bdy_index(i),:));
end
bdd_len = bdd_len + norm(v(bdy_index(1),:)-v(bdy_index(nb),:));
len_c = norm(v(bdy_index(1),:)-v(bdy_index(nb),:));
for i = 1:nb
    diskbdy_x(i,1) = cos(len_c/bdd_len*2*pi);
    diskbdy_y(i,1) = sin(len_c/bdd_len*2*pi);
    if i < nb
        len_c = len_c + norm(v(bdy_index(i+1),:)-v(bdy_index(i),:));
    end
end

% map to disk
disk = pc_disk_harmonic_map(v,bdy_index,[diskbdy_x,diskbdy_y],patch);

%% Step 2: Mapping the unit disk to the unit square

% compute the generalized Laplacian
patch = calc_patch_polyfit(disk,patch);
mu = calc_bc_2dto3d(disk,v,patch);

% set the boundary constraints
bottom = bdy_index(id1:id2);
right = bdy_index(id2:id3);
top = bdy_index(id3:id4);
left = bdy_index([id4:end,id1]);

Ax = Lap_nn(mu,patch); Ay = Ax;
bx = zeros(nv,1); by = bx;
Ax([left;right],:) = 0;
Ax([left;right],[left;right]) = diag(ones(length([left;right]),1));
bx(right) = 1;
Ay([top;bottom],:) = 0;
Ay([top;bottom],[top;bottom]) = diag(ones(length([top;bottom]),1));
by(top) = 1;

% solve the generalized Laplace equation
square_x = Ax\bx;
square_y = Ay\by;

%% Step 3: Optimize the height of the square to achieve a conformal map
 
h_opt = fminbnd(@(h) h_energy(h,square_x,square_y,0,patch),0,10);
map = [square_x, h_opt*square_y];

map_mu = calc_bc_2dto3d(map,v,patch);

end

function E = h_energy(h,sol_u,sol_v,mu,patch)
    h_mu = calc_pcbc([sol_u h*sol_v],patch);
    E = sqrt(sum(abs(h_mu(:,1)-mu).^2));
end

function mu = calc_pcbc(fz,patch)
    [n,k] = size(patch.knnInd);

    C = patch.C;
    knnInd = patch.knnInd;
    coord = patch.coord;

    c2f1 = sum(C(:, 2:6:6*k-4) .* fz(knnInd), 2);
    c4f1 = sum(C(:, 4:6:6*k-2) .* fz(knnInd), 2);
    c5f1 = sum(C(:, 5:6:6*k-1) .* fz(knnInd), 2);
    ux = repmat(c2f1, [1,k]) + 2 .* coord(:, 1:k) .* repmat(c4f1, [1,k]) + coord(:, k+1:2*k) .* repmat(c5f1, [1,k]);

    c2f2 = sum(C(:, 2:6:6*k-4) .* fz(knnInd+n), 2);
    c4f2 = sum(C(:, 4:6:6*k-2) .* fz(knnInd+n), 2);
    c5f2 = sum(C(:, 5:6:6*k-1) .* fz(knnInd+n), 2);
    vx = repmat(c2f2, [1,k]) + 2 .* coord(:, 1:k) .* repmat(c4f2, [1,k]) + coord(:, k+1:2*k) .* repmat(c5f2, [1,k]);

    c3f1 = sum(C(:, 3:6:6*k-3) .* fz(knnInd), 2);
    c6f1 = sum(C(:, 6:6:6*k) .* fz(knnInd), 2);
    uy = repmat(c3f1, [1,k]) + 2 .* coord(:, k+1:2*k) .* repmat(c6f1, [1,k]) + coord(:, 1:k) .* repmat(c5f1, [1,k]);

    c3f2 = sum(C(:, 3:6:6*k-3) .* fz(knnInd+n), 2);
    c6f2 = sum(C(:, 6:6:6*k) .* fz(knnInd+n), 2);
    vy = repmat(c3f2, [1,k]) + 2 .* coord(:, k+1:2*k) .* repmat(c6f2, [1,k]) + coord(:, 1:k) .* repmat(c5f2, [1,k]);

    E = ux.^2 + vx.^2;
    F = ux.*uy + vx.*vy;
    G = uy.^2 + vy.^2;

    D = E + G + 2*sqrt(E.*G - F.^2);
    mu = ( E - G + 2i * F) ./ D;
end

function A = Lap_nn(mu, patch)
    af = (1-2*real(mu)+abs(mu).^2)./(1.0-abs(mu).^2);
    bf = -2*imag(mu)./(1.0-abs(mu).^2);
    gf = (1+2*real(mu)+abs(mu).^2)./(1.0-abs(mu).^2);

    [n,k] = size(patch.knnInd);
    D = patch.Differential;
    knnInd = patch.knnInd;

    af_all = af(knnInd);
    bf_all = bf(knnInd);
    gf_all = gf(knnInd);

    p1 = repmat(af_all(:,1), [1,k]) .* D.dxx + 2 .* repmat(bf_all(:,1), [1,k]) .* D.dxy + repmat(gf_all(:,1), [1,k]) .* D.dyy;
    ax = sum(D.dx .* af_all, 2);
    bx = sum(D.dx .* bf_all, 2);
    by = sum(D.dy .* bf_all, 2);
    gy = sum(D.dy .* gf_all, 2);
    p2 = repmat(ax, [1,k]) .* D.dx + repmat(bx, [1,k]) .* D.dy + repmat(by, [1,k]) .* D.dx + repmat(gy, [1,k]) .* D.dy;
    V = p1+p2;
    I = repmat((1:n)', [1, k]);
    J = knnInd;

    A = sparse(I,J,-V);

    A_bdd = Lap_nn_bddlm_slow(mu, patch);
    A(patch.bdd_index, :) = 0;
    A = A + A_bdd;
end

function A = Lap_nn_bddlm_slow(mu, patch)
    index = patch.bdd_index;
    k = patch.k;
    I = 0.* patch.onering(index,:);
    J = 0.* patch.onering(index,:);
    V = 0.* patch.onering(index,:);

    for i=1:size(index,1)
        index_ = index(i);
        x = patch.coord(index_, 1:k).';
        y = patch.coord(index_, k+1:2*k).';
        TRI = delaunay(x,y);

        delete_ind = [];
        knnInd_local = patch.knnInd(index_,:)';
        for j=1:size(TRI,1)
            if ismember(knnInd_local(TRI(j,:)), index)
                delete_ind = [delete_ind; j];
            end
        end
        if length(delete_ind) < size(TRI,1)
            TRI(delete_ind,:) = [];
        end


        nbhd = patch.knnInd(index_, :).';
        mu_pt = mu(nbhd, 1);
        mu_ = (mu_pt(TRI(:,1),1) +mu_pt(TRI(:,2),1) +mu_pt(TRI(:,3),1))./3;
        af = (1-2*real(mu_)+abs(mu_).^2)./(1.0-abs(mu_).^2);
        bf = -2*imag(mu_)./(1.0-abs(mu_).^2);
        gf = (1+2*real(mu_)+abs(mu_).^2)./(1.0-abs(mu_).^2);

        f0 = TRI(:,1);
        f1 = TRI(:,2);
        f2 = TRI(:,3);

        uxv0 = y(f1) - y(f2);
        uyv0 = x(f2) - x(f1);
        uxv1 = y(f2) - y(f0);
        uyv1 = x(f0) - x(f2);
        uxv2 = y(f0) - y(f1);
        uyv2 = x(f1) - x(f0);

        l = [sqrt(sum(uxv0.^2 + uyv0.^2,2)) ...
            sqrt(sum(uxv1.^2 + uyv1.^2,2)) ...
            sqrt(sum(uxv2.^2 + uyv2.^2,2))];
        s = sum(l,2)*0.5;

        area = sqrt( s.*(s-l(:,1)).*(s-l(:,2)).*(s-l(:,3)));

        v00 = (af.*uxv0.*uxv0 + 2*bf.*uxv0.*uyv0 + gf.*uyv0.*uyv0)./area;
        v11 = (af.*uxv1.*uxv1 + 2*bf.*uxv1.*uyv1 + gf.*uyv1.*uyv1)./area;
        v22 = (af.*uxv2.*uxv2 + 2*bf.*uxv2.*uyv2 + gf.*uyv2.*uyv2)./area;

        v01 = (af.*uxv1.*uxv0 + bf.*uxv1.*uyv0 + bf.*uxv0.*uyv1 + gf.*uyv1.*uyv0)./area;
        v12 = (af.*uxv2.*uxv1 + bf.*uxv2.*uyv1 + bf.*uxv1.*uyv2 + gf.*uyv2.*uyv1)./area;
        v20 = (af.*uxv0.*uxv2 + bf.*uxv0.*uyv2 + bf.*uxv2.*uyv0 + gf.*uyv0.*uyv2)./area;

        I_ = [f0;f1;f2;f0;f1;f1;f2;f2;f0];
        J_ = [f0;f1;f2;f1;f0;f2;f1;f0;f2];
        V_ = [v00;v11;v22;v01;v01;v12;v12;v20;v20]/2;

        J_(I_~=1) = [];
        V_(I_~=1) = [];
        I_(I_~=1) = [];
        I_ = nbhd(I_);
        J_ = nbhd(J_);

        I(i,1:numel(I_)) = I_;
        J(i,1:numel(I_)) = J_;
        V(i,1:numel(I_)) = V_;
    end

    J(I==0)=[];
    V(I==0)=[];
    I(I==0)=[];

    A = sparse(I, J, V, size(mu,1), size(mu,1));
end



