% Compute the point cloud rectangular conformal mapping using the fast method in [1].
%
% Usage:
% [map,map_mu] = pc_rectangular_conformal_map(v,bdy_index,corner)
% 
% Input:
% v: nv x 3 vertex coordinates of a disk-shaped point cloud
% bdy_index: nb x 1 vertex indices of the boundary, with anti-clockwise orientation
% (optional) corner: 4 x 1 vertex indices for the four corners of the rectangle, with anti-clockwise orientation
% 4 - 3
% |   |
% 1 - 2
%
% Output:
% map: nv x 2 vertex coordinates of the rectangular conformal parameterization
% map_mu: nv x 1 Beltrami coefficient whose norm indicates the conformality distortion
%
% If you use this code in your own work, please cite the following paper:
% [1] T. W. Meng, G. P.-T. Choi and L. M. Lui, 
%     "TEMPO: Feature-Endowed Teichmller Extremal Mappings of Point Clouds."
%     SIAM Journal on Imaging Sciences, 9(4), pp. 1922-1962, 2016.
%
% Copyright (c) 2015-2018, Gary Pui-Tung Choi
% https://scholar.harvard.edu/choi

addpath('code')

%% Example 1: human face
load human_face.mat

% plot the point cloud
figure;
plot3(v(:,1),v(:,2),v(:,3),'.'); 
axis equal tight
title('The input point cloud')
set(gca,'FontSize',12);
view([-10 70])

% Compute the rectangular conformal parameterization of the point cloud
tic;
[map,map_mu] = pc_rectangular_conformal_map(v,bdyid);
toc;

% plot the parameterization result
figure;
plot(map(:,1),map(:,2),'.'); 
axis equal tight
title('Rectangular conformal parameterization')
set(gca,'FontSize',12);

% histogram plot of the conformality distortion of the parameterization
% should be very close to 0
figure;
hist(abs(map_mu),0:0.001:1);
title('Conformality distortion')
xlabel('Norm of Beltrami coefficient')
ylabel('Number of vertices')
set(gca,'FontSize',12);

%% Example 2: spiral
load spiral.mat

% plot the point cloud
figure;
plot3(v(:,1),v(:,2),v(:,3),'.'); 
axis equal tight
title('The input point cloud')
set(gca,'FontSize',12);
view([90 -80])

% Compute the rectangular conformal parameterization of the point cloud
tic;
[map,map_mu] = pc_rectangular_conformal_map(v,bdyid);
toc;

% plot the parameterization result
figure;
plot(map(:,1),map(:,2),'.'); 
axis equal tight
title('Rectangular conformal parameterization')
set(gca,'FontSize',12);

% histogram plot of the conformality distortion of the parameterization
% should be very close to 0
figure;
hist(abs(map_mu),0:0.001:1);
title('Conformality distortion')
xlabel('Norm of Beltrami coefficient')
ylabel('Number of vertices')
set(gca,'FontSize',12);
